
const uint16_t NUMCHARS = 4096; // framed packet maximum size, plenty big to accomodate bit errors.
const char EOFCHAR = '\r';
//extern const bool DEBUG;

// Types of sentences as defined by the Dynon/Garmin protocol. For realtime purposes, enable only the
// sentences actually required, otherwise buffer overflow may occur.
//
const char ADAHRS_TYPE = '1';
const char SYSTEM_TYPE = '2'; // Dynon Skyview/AFS only)
const char EMS_TYPE = '3';

const char DYNON = '!';
const uint16_t dADAHRS = 74;
const uint16_t dSYSTEM = 93;
const uint16_t dEMS = 225;
// Internal Packet Identifier
const uint8_t ID_SV_ADAHRS = 1;
const uint8_t ID_SV_SYSTEM = 2;
const uint8_t ID_SV_EMS = 3;

// Engine variables
extern int16_t OilP; // -100  invalid
extern int16_t OilT; // -1000 invalid
extern uint16_t RpmL;
extern uint16_t RpmR;
extern int16_t Map;     // -100  invalid
extern int16_t FuelF1;  // -100  invalid
extern int16_t FuelF2;  // -100  invalid
extern int16_t FuelP;   // -100  invalid
extern int16_t FuelL;   // -100  invalid
extern int16_t FuelR;   // -100  invalid
extern int16_t FuelRem; // -100  invalid
extern int16_t Volts1;  // -100  invalid
extern int16_t Volts2;  // -100  invalid
extern int16_t Amps;    // -1000 invalid
extern uint32_t HobbsT;
extern uint32_t TachT;
extern int16_t TCpl[14];   // -1000 invalid
extern int16_t GpIn[13];   // -10000invalid
extern int8_t CtctIn[16];  // -1    invalid
extern int16_t PercentPwr; // -100  invalid
extern char LeanState;
// ADARHS variables
extern int16_t  Pitch;
extern int16_t  Roll;
extern int16_t  Heading;   
extern int16_t  AirSpeed;
extern int32_t  PressAlt;
extern int32_t  Altitude;  
extern int16_t  TurnRate;
extern int16_t  LatAccel;
extern int16_t  VertAccel ;
extern int16_t  AOA;
extern int16_t  VertSpeed;
extern int16_t  OAT;
extern int16_t  TASpeed;  
extern int16_t  Baro      ;//= 242; 
extern int32_t  DensAlt;     
extern int16_t  WindDir   ;//= 0; 
extern int16_t  WindSpeed ;//= 0;
// System variables (Dynon Only)
extern int16_t  HdgBug     ;//= -1;
extern int16_t  AltBug     ;//= -10000;
extern int16_t  ASBug      ;//= -1;
extern int16_t  VSBug      ;//= -1000;
extern int16_t  Course     ;//= -1;
extern uint8_t  CDSTyp;
extern uint8_t  CDSPor;
extern uint8_t  CDScal;
extern int16_t  CDDefl     ;//= -100;
extern int16_t  GlideS     ;//= -100;
extern uint8_t  APEngd;
extern uint8_t  APRMod;
extern uint8_t  APPMod;
extern int16_t  APRFor;
extern int16_t  APRPos    ;//= -10000;
extern uint8_t  APRSlp;   ;//
extern int16_t  APPFor;   ;//
extern int16_t  APPPos    ;//= -10000;
extern uint8_t  APPSlp;   ;//
extern int16_t  APYFor;   ;//
extern int16_t  APYPos    ;//= -10000;
extern uint8_t  APYSlp;
extern uint8_t  XPStat;
extern uint8_t  XPRep;
extern uint8_t  XPID;
extern uint16_t XPCode    ;//= 00000;

extern int16_t Slip;
extern float   SmoothedLateralG;
extern const float slipSmoothingAlpha; // = .5; //1 = max smoothing, 0.01 no smoothing.
extern uint16_t 	   ballSize; // necessary for Slip Filter
extern int16_t  FlightPath		;//= -1000;
extern int16_t  GroundTrack		;//= -1000;
extern int16_t  GroundSpeed     ;//= -1;
extern int16_t  DriftAngle      ;//= -1000;


//
// Method Prototypes
//
void serPacketFrame(int16_t serialPort);
void udpPacketFrame(char *buffer, uint16_t len);
uint16_t sentenceDecode(char *buffer, uint16_t start, uint16_t len);
int32_t extractField(char *buffer, uint16_t start, uint16_t len, int32_t nodata);
int32_t getSignedInt32(char *buffer, uint16_t start, uint8_t len);
// uint32_t extractASCIIHex32(char *buffer, uint16_t start, uint8_t len);

class EFISdata
{
private:
public:
  char packetChars[NUMCHARS]; // frame aligned receive buffer
  uint16_t packetLen;         // adjusted packet length
  bool packetValid = false;   // indicates receive buffer state
  uint8_t PacketID;
  uint16_t checkSumErrs;

  uint16_t charIndex;
  uint16_t checkSum;
  char rxChar;
  uint16_t offSet = 0;

  void serPacketFrame(int16_t serialPort) {
    if (rxChar != '\n')
    {
      checkSum += rxChar; // update checksum
      packetChars[charIndex] = rxChar;

      if (rxChar == EOFCHAR)
      {
        //
        // Process checksum by removing last 3 characters from computation (transmitted checksum and cr)
        //
        checkSum = (checkSum - packetChars[charIndex] - packetChars[charIndex - 1] - packetChars[charIndex - 2]) % 256;

        // convert ASCII checksum to byte value
        byte checkSumRx = packetChars[charIndex - 2] - '0';
        if (checkSumRx > 9)
          checkSumRx -= 7;
        byte checkSumRx2 = packetChars[charIndex - 1] - '0';
        if (checkSumRx2 > 9)
          checkSumRx2 -= 7;

        checkSumRx = (checkSumRx << 4) + checkSumRx2;
        /*
        if (DEBUG)
          Serial.print("Checksum Received: ");
        if (DEBUG)
          Serial.print(checkSumRx, HEX);
        if (DEBUG)
          Serial.print(" Checksum Computed: ");
        if (DEBUG)
          Serial.println(checkSum, HEX);
        */
        boolean checkSumMatch = (checkSumRx == checkSum);

        // count checksum errors for higher level processing
        if (!checkSumMatch)
          checkSumErrs++;

        // Jam in a linefeed and null to make a standard packet format.
        // The null makes it printable.
        charIndex++;
        packetChars[charIndex] = '\n';
        packetChars[charIndex + 1] = '\0';
        /*
        if (DEBUG)
          Serial.println(packetChars);
          */
        packetLen = charIndex + 1;
        checkSum = 0;
        charIndex = 0;      // ready for next sentence
        packetValid = true; // maintain sentence for udp transmission

        char vendorType = packetChars[0];
        char packetType = packetChars[1];

        //
        // Process a valid sentence. Demo modes disable checksum validation.
        //
        if (checkSumMatch)
        {
          // Check the sentence credentials (vendor and type) and match against the expected sentence length.
          // For future enhancements: this is where sentence recognition occurs.
          if (vendorType == DYNON && packetType == ADAHRS_TYPE && packetLen == dADAHRS) extract_dADAHRS(offSet);
          // if (vendorType == DYNON && packetType == SYSTEM_TYPE && packetLen == dSYSTEM)
          // extract_dSYSTEM(offSet);
          if (vendorType == DYNON && packetType == EMS_TYPE && packetLen == dEMS) extract_dEMS(offSet);
        } // if checkSumMatch
      }   // if EOF

      else
        charIndex++;
      if (charIndex >= NUMCHARS)
        charIndex = NUMCHARS - 3; // saturate buffer, allowing room for lf/null end tags.
    }                             // if (rxChar != '\n')
  }                               // serPacketFrame()

  //
  // Extract fields from a valid Dynon EMS packet.
  void extract_dEMS(uint16_t offSet)
  {
    OilP = extractField(packetChars, offSet + 11, 3, -1);
    OilT = extractField(packetChars, offSet + 14, 4, -1);
    RpmL = extractField(packetChars, offSet + 18, 4, -1);
    RpmR = extractField(packetChars, offSet + 22, 4, -1);
    Map = extractField(packetChars, offSet + 26, 3, -10);
    FuelF1 = extractField(packetChars, offSet + 29, 3, -10);
    FuelF2 = extractField(packetChars, offSet + 32, 3, -10);
    FuelP = extractField(packetChars, offSet + 35, 3, -10);
    FuelL = extractField(packetChars, offSet + 38, 3, -10);
    FuelR = extractField(packetChars, offSet + 41, 3, -10);
    FuelRem = extractField(packetChars, offSet + 44, 3, -10);
    Volts1 = extractField(packetChars, offSet + 47, 3, -10);
    Volts2 = extractField(packetChars, offSet + 50, 3, -10);
    Amps = extractField(packetChars, offSet + 53, 4, -1000);
    HobbsT = extractField(packetChars, offSet + 57, 5, -1);
    TachT = extractField(packetChars, offSet + 62, 5, -1);

    for (uint8_t i = 0; i < 14; i++)
    {
      TCpl[i] = getSignedInt32(packetChars, offSet + 67 + i * 4, 4);
    }

    for (uint8_t i = 0; i < 13; i++)
    {
      GpIn[i] = getSignedInt32(packetChars, offSet + 123 + i * 6, 5);
    }

    for (uint8_t i = 0; i < 16; i++)
    {
      char startChar = packetChars[offSet + 201 + i];
      if (startChar == '0' || startChar == 'V')
        CtctIn[i] = 0;
      else if (startChar == '1' || startChar == 'A')
        CtctIn[i] = 1;
      else
        CtctIn[i] = -1;
    }

    PercentPwr = extractField(packetChars, offSet + 217, 3, -100); //-1
    LeanState = packetChars[offSet + 220];

    PacketID = ID_SV_EMS;
  }

  // Extract fields from a valid Dynon, Garmin, or OnSpeed packets.
void extract_dADAHRS(uint16_t offSet)
{
  Pitch     		= extractField (packetChars, offSet + 11, 4, -1000);
  Roll      		= extractField (packetChars, offSet + 15, 5, -10000);
  Heading   		= extractField (packetChars, offSet + 20, 3, -1);
  AirSpeed  		= extractField (packetChars, offSet + 23, 4, -1);
  PressAlt  		= extractField (packetChars, offSet + 27, 6, -10000);
  TurnRate  		= extractField (packetChars, offSet + 33, 4, -1000);
  LatAccel  		= extractField (packetChars, offSet + 37, 3, -100);
  VertAccel 		= extractField (packetChars, offSet + 40, 3, -100);
  AOA       		= extractField (packetChars, offSet + 43, 2, -1);
  VertSpeed 		= extractField (packetChars, offSet + 45, 4, -1000);
  OAT       		= extractField (packetChars, offSet + 49, 3, -100);
  TASpeed   		= extractField (packetChars, offSet + 52, 4, -1);
  Baro      		= extractField (packetChars, offSet + 56, 3, 242);
  DensAlt   		= extractField (packetChars, offSet + 59, 6, -10000);
  WindDir   		= extractField (packetChars, offSet + 65, 3, 0);
  WindSpeed 		= extractField (packetChars, offSet + 68, 2, 0);

  // Computed indicated altitude
  Altitude  		= (int32_t)(PressAlt - (145366.45f * (1 - pow ((Baro + 2750) / 2992.126f,  0.190284f))) + 2);
  if (Altitude < -10000) 	Altitude = -10000;
  if (Altitude >= 100000) 	Altitude = -10000;

  SmoothedLateralG 	= SmoothedLateralG * slipSmoothingAlpha + (1 - slipSmoothingAlpha) * (float)LatAccel * 0.01f;
  // 6.39245322f = 1.5 degrees per ball, 12.7849064f = 3 degrees per ball.
  Slip = int16_t (SmoothedLateralG * ballSize * 12.7849064f);
  Slip = constrain(Slip, -99, 99);

  FlightPath  		= atan2 (VertSpeed * 10, TASpeed * 10.12685f) * RAD_TO_DEG * 10; // degrees * 10
  int16_t px0 		= TASpeed * cos (Heading * DEG_TO_RAD) + WindSpeed * 10 * cos (WindDir * DEG_TO_RAD);
  int16_t py0 		= TASpeed * sin (Heading * DEG_TO_RAD) + WindSpeed * 10 * sin (WindDir * DEG_TO_RAD);

  GroundTrack 		= atan2 (py0, px0) * RAD_TO_DEG;
  if (GroundTrack < 0) GroundTrack += 360; // convert to standard heading

  GroundSpeed 		= sqrt (px0 * px0 + py0 * py0); // in deciknots
  DriftAngle  		= Heading - GroundTrack; // in whole degrees

  PacketID  = ID_SV_ADAHRS;
}

  // When a UDP packet is received, assume it is well-formed.
  // A standard Dynon UDP packet will have three sentences:
  // ADAHRS, SYSTEM and EMS, each starting with an '!' and
  // terminated with a \r\n.
  //
  void udpPacketFrame(char *packetChars, uint16_t packetLen) {

    int16_t sentenceLen;    // expected length of the sentence
    uint16_t sentenceCheck; // actual length of the sentence after decoding, or 0 if checksum fails
    uint16_t offSet;        // pointer to the start of the sentence to be decoded

    char rxChar = packetChars[0];
    char rxChar1 = packetChars[1];

    //
    // Check for packet triplet as sent by Dynon SkyView only.
    //
    //if (rxChar == DYNON && rxChar1 == ADAHRS_TYPE && packetLen == (dADAHRS + dSYSTEM + dEMS)) {
      offSet = 0;
      sentenceLen = dADAHRS;
      sentenceCheck = sentenceDecode(packetChars, offSet, sentenceLen);
      if (!sentenceCheck)
        checkSumErrs++;

      if (sentenceCheck == dADAHRS){
        extract_dADAHRS(offSet);
        packetValid = true;
      }

      offSet = dADAHRS;
      sentenceLen = dSYSTEM;
      sentenceCheck = sentenceDecode(packetChars, offSet, sentenceLen);
      if (!sentenceCheck)
        checkSumErrs++;

      if (sentenceCheck == dSYSTEM) {
        //extract_dSYSTEM(offSet);
        packetValid = true;
      }

      offSet = dADAHRS + dSYSTEM;
      sentenceLen = dEMS;
      sentenceCheck = sentenceDecode(packetChars, offSet, sentenceLen);
      if (!sentenceCheck)
        checkSumErrs++;

      if (sentenceCheck == dEMS) {
        extract_dEMS(offSet);
        packetValid = true;
      }
    //} // if DYNON

  } // udpPacketFrame

  // Decode sentences from the buffer position <start> to the expected length <len>
  // and returns the actual length. A \n newline character will terminate the decoding
  // early. The calling program should check the returned length to validate the sentence.
  //
  uint16_t sentenceDecode(char *packetChars, uint16_t start, uint16_t len)
  {

    byte checkSum = 0;
    uint16_t index;
    char rxChar;

    //
    // This loop can terminate early if a newline character is detected.
    // Otherwise, it uses the length parameter to terminate.
    //
    for (index = start; index < (start + len); index++)
    {
      rxChar = packetChars[index];
      checkSum += rxChar;
      //if (DEBUG)
      //  Serial.print(rxChar);
      if (rxChar == '\n')
        break;
    }

    // checksum can be corrected by subtraction of checksum and delimiter characters
    // <LF><packet data><Chk1><Chk2><CR>
    checkSum = (checkSum - packetChars[index] - packetChars[index - 1] -
                packetChars[index - 2] - packetChars[index - 3]) %
               256;
    byte checkSumRx = packetChars[index - 3] - '0';
    if (checkSumRx > 9)
      checkSumRx -= 7;
    byte checkSumRx2 = packetChars[index - 2] - '0';
    if (checkSumRx2 > 9)
      checkSumRx2 -= 7;
    checkSumRx = (checkSumRx << 4) + checkSumRx2;
    /*
    if (DEBUG)
      Serial.print(" Checksum Received: ");
    if (DEBUG)
      Serial.print(checkSumRx, HEX);
    if (DEBUG)
      Serial.print(" Checksum Computed: ");
    if (DEBUG)
      Serial.println(checkSum, HEX);
    */
    return ((checkSumRx == checkSum) ? index - start + 1 : 0);
  } // ends sentenceDecode

  //
  // Extract data from 'buffer' and return with result.  Returns 'nodata' value if unprovisioned.
  //
  int32_t extractField(char *buffer, uint16_t start, uint16_t len, int32_t nodata)
  {
    char startChar = buffer[start];
    if (startChar == 'X' || startChar == 'Z' || startChar == '_')
      return nodata;
    else
      return getSignedInt32(buffer, start, len);
  }

  //
  // Extract a signed or unsigned integer value from a character array 'buffer'
  // starting at position 'start' until length 'len'.
  // Illegal numerical character values are forced to 0.
  //
  int32_t getSignedInt32(char buffer[], uint16_t start, uint8_t len)
  {
    int32_t Value = 0;
    uint8_t digitV = 0;

    if ((start + len) > (NUMCHARS - 1))
      return 0;

    if (buffer[start] == '-')
    {
      for (uint8_t i = 1; i < len; i++)
      {
        digitV = buffer[start + i] - '0';
        if (digitV > 9)
          digitV = 0; // errors default to 0
        Value = 10 * Value + (digitV);
      }
      return -Value;
    }

    else if (buffer[start] == '+')
    {
      for (uint8_t i = 1; i < len; i++)
      {
        digitV = buffer[start + i] - '0';
        if (digitV > 9)
          digitV = 0; // errors default to 0
        Value = 10 * Value + (digitV);
      }
      return Value;
    }

    else
    {
      for (uint8_t i = 0; i < len; i++)
      {
        digitV = buffer[start + i] - '0';
        if (digitV > 9)
          digitV = 0; // errors default to 0
        Value = 10 * Value + (digitV);
      }
      return Value;
    }
  }

}; // ends EFISdata class function


